/*
 * Copyright (C) 2017  Grandstream Networks, Inc.
 *
 * Author:  <hz_sys@grandstream.cn>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/cdev.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/io.h>
#include <linux/sched.h>
#include <linux/wait.h>
#include <linux/miscdevice.h>
#include <linux/poll.h>
#include <linux/interrupt.h>
#include <../../misc/mediatek/include/mt-plat/mt_gpt.h>

#ifdef CONFIG_OF
#include <linux/of_irq.h>
#include <linux/of.h>
#include <linux/of_address.h>
#endif

#define AVS_DEV_MAX 1

#define AVS_TIMER_START 0x01
#define AVS_TIMER_STOP 0x02
#define AVS_TIMER_SET_INTERVAL 0x03

#if 0
extern int stop_gpt(unsigned int id);
extern int start_gpt(unsigned int id);
extern int gpt_set_cmp(unsigned int id, unsigned int val);
#endif

struct avs_timer {
	struct miscdevice miscdev;
	int irq;
	unsigned int interval;
	bool done;
	unsigned long dev_opened;
	wait_queue_head_t wq;
	struct tasklet_struct tasklet;
	spinlock_t lock;
};

struct avs_timer g_timer;

static void timer_set_interval(struct avs_timer *timer, u64 us)
{
	printk ("avs timer set %llu us\n", us);
	if (us >= 1000 && us <=1000000) /* 1ms ~ 1s */
		gpt_set_cmp(GPT5, (us/1000) * 13000);
}

static inline void timer_enable(void)
{
	printk ("avs_timer start\n");
	start_gpt (GPT5);
}

static inline void timer_disable(void)
{
	printk ("avs_timer stop\n");
	stop_gpt (GPT5);
}

#if 0
static irqreturn_t timer_isr(int irq, void *dev)
{
	//unsigned int reg_val;
	struct avs_timer *timer = (struct avs_timer *)dev;

#if 0
	/* clr timer4 interrupt status */
	reg_val = (1 << 4);
	writel(reg_val, TMR_IRQ_STA_REG);
#endif
	printk ("avs_timer interrupt\n");
	tasklet_schedule(&timer->tasklet);
	return IRQ_HANDLED;
}
#endif

void gpt_handler (unsigned long data)
{
	//struct avs_timer *timer = (struct avs_timer *)data;
	tasklet_schedule(&g_timer.tasklet);
}

static void avs_timer_tasklet(unsigned long data)
{
	struct avs_timer *timer = (struct avs_timer *)data;
	unsigned long flags;

	spin_lock_irqsave(&timer->lock, flags);
	timer->done = true;
	wake_up(&timer->wq);
	spin_unlock_irqrestore(&timer->lock, flags);
}

static int avs_timer_open(struct inode *inode, struct file *file)
{
	struct avs_timer *timer = container_of(file->private_data,
					       struct avs_timer, miscdev);

	if (test_and_set_bit(0, &timer->dev_opened))
		return -EBUSY;

	return 0;
}

static int avs_timer_release(struct inode *inode, struct file *file)
{
	struct avs_timer *timer = container_of(file->private_data,
					       struct avs_timer, miscdev);

	clear_bit(0, &timer->dev_opened);
	timer_disable();

	return 0;
}

static unsigned int avs_timer_poll(struct file *file, poll_table *wait)
{
	struct avs_timer *timer = container_of(file->private_data,
					       struct avs_timer, miscdev);
	unsigned int ret;
	unsigned long flags;

	poll_wait(file, &timer->wq, wait);

	spin_lock_irqsave(&timer->lock, flags);
	if (timer->done) {
		ret = POLLIN | POLLRDNORM;
		timer->done = false;
	} else
		ret = 0;
	spin_unlock_irqrestore(&timer->lock, flags);

	return ret;
}

static long avs_timer_ioctl(struct file *file, unsigned int cmd,
			    unsigned long arg)
{
	struct avs_timer *timer = container_of(file->private_data,
					       struct avs_timer, miscdev);

	switch (cmd) {
	case AVS_TIMER_START:
		timer_enable();
		break;
	case AVS_TIMER_STOP:
		timer_disable();
		break;
	case AVS_TIMER_SET_INTERVAL:
		{
			unsigned int val;

			if (copy_from_user(&val, (void __user *)arg, sizeof(unsigned int)))
				return -EFAULT;

			timer->interval = val;
			timer_set_interval(timer, timer->interval);
			break;
		}
	default:
		return -ENOTTY;
	}

	return 0;
}

static const struct file_operations avs_timer_fops = {
	.owner = THIS_MODULE,
	.llseek = no_llseek,
	.open = avs_timer_open,
	.release = avs_timer_release,
	.poll = avs_timer_poll,
	.unlocked_ioctl = avs_timer_ioctl,
};

static int avs_timer_probe(struct platform_device *pdev)
{
	struct device *dev = &pdev->dev;
	struct avs_timer *timer;
	//struct device_node *node = NULL;
	int err;

	timer = kzalloc(sizeof(struct avs_timer), GFP_KERNEL);
	if (timer == NULL) {
		err = -ENOMEM;
		goto err_return;
	}

	spin_lock_init(&timer->lock);
	tasklet_init(&timer->tasklet, avs_timer_tasklet, (unsigned long)timer);
	init_waitqueue_head(&timer->wq);

	timer->miscdev.minor = MISC_DYNAMIC_MINOR;
	timer->miscdev.name = "avs_timer";
	timer->miscdev.fops = &avs_timer_fops;
	err = misc_register(&timer->miscdev);
	if (err) {
		dev_err(dev, "Register %s failed\n", timer->miscdev.name);
		goto err_free_priv;
	}
#if 0
#ifdef CONFIG_OF
	node = of_find_compatible_node(NULL, NULL, "mediatek,mt8163-avsgpt");
	if (node == NULL)
		pr_debug("avsgpt get node failed\n");

	/*get irq num */
	timer->irq = irq_of_parse_and_map(node, 0);;
#else
	timer->irq = platform_get_irq(pdev, 0);
#endif
	if (timer->irq <= 0) {
		dev_err(dev, "avs_timer failed to attach irq number\n");
		goto err_misc_register;
	}
	err = request_irq(timer->irq, timer_isr,
			  IRQF_DISABLED, "avs_timer", timer);
	if (err) {
		err = -ENODEV;
		dev_err(dev, "avs_timer failed to request timer irq\n");
		goto err_misc_register;
	}
#endif

	err = request_gpt(GPT5, GPT_REPEAT, GPT_CLK_SRC_SYS,
			GPT_CLK_DIV_1, 0, gpt_handler, GPT_NOAUTOEN);        
	if (err) {
		printk("[%s]fail to request GPT%d\n", __func__, GPT5 + 1);
		goto err_misc_register;
	}

	gpt_set_cmp(GPT5, 130000); // 10ms
	timer->interval = 10000; //default 10ms

	platform_set_drvdata(pdev, timer);
	g_timer = *timer;

	return 0;
err_misc_register:
	misc_deregister(&timer->miscdev);
err_free_priv:
	kfree(timer);
err_return:
	return err;
}

#if 0
static int avs_timer_remove(struct platform_device *pdev)
{
	struct avs_timer *timer = dev_get_drvdata(&pdev->dev);

	free_irq(timer->irq, timer);
	misc_deregister(&timer->miscdev);
	dev_set_drvdata(&pdev->dev, NULL);
	kfree(timer);

	return 0;
}
#endif

#if defined CONFIG_PM
int avs_timer_suspend(struct device *dev)
{

	//disable_irq(timer->irq);
	timer_disable();
	return 0;
}

int avs_timer_resume(struct device *dev)
{
	struct avs_timer *timer = dev_get_drvdata(dev);

	timer_set_interval(timer, timer->interval);
	if (timer->dev_opened)
		timer_enable();
	//enable_irq(timer->irq);
	return 0;
}

static const struct dev_pm_ops avs_timer_pm_ops = {
	.suspend = avs_timer_suspend,
	.resume = avs_timer_resume,
};

//#else
//static const struct dev_pm_ops avs_timer_fops;
#endif

#ifdef CONFIG_OF
static const struct of_device_id avs_timer_of_ids[] = {                                                                  
	{.compatible = "mediatek,mt8163-avsgpt",},
	{}
};
#endif

static struct platform_driver avs_timer_driver = {
	.probe = avs_timer_probe,
//	.remove = __devexit_p(avs_timer_remove),
	.driver = {
		   .name = "avs_timer",
		   .owner = THIS_MODULE,
#if defined CONFIG_PM
		   .pm = &avs_timer_pm_ops,
#endif
#ifdef CONFIG_OF
		   .of_match_table = avs_timer_of_ids,
#endif
		   },
};

module_platform_driver(avs_timer_driver);
