/*
 * Copyright (C) 2017  Grandstream Networks, Inc.
 * Author: Fanan Chen <fnchen@grandstream.cn>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/gpio.h>
#include <linux/interrupt.h>
#include <linux/of.h>
#include <linux/of_gpio.h>
#include <linux/of_platform.h>
#include <linux/platform_device.h>
#include <linux/slab.h>
#include <linux/workqueue.h>
#include <linux/switch.h>

#define POLL_TIME 100

struct gs_switch_data {
	u32 channel_id;
	u32 detect_voltage[2];
	int state;
	struct delayed_work dwork;
	struct switch_dev sdev;
};

struct gs_switch_priv {
	int num_devs;
	struct gs_switch_data devs[];
};

extern int IMM_GetOneChannelValue_Cali(int Channel, int *voltage);

static void gs_switch_work(struct work_struct *work)
{
	int voltage, state;
	struct gs_switch_data *gs_switch =
		container_of(work, struct gs_switch_data, dwork.work);

	IMM_GetOneChannelValue_Cali(gs_switch->channel_id, &voltage);
	voltage = voltage / 1000;

	if (voltage > gs_switch->detect_voltage[0] &&
	    voltage < gs_switch->detect_voltage[1])
		state = 1;
	else
		state = 0;

	if (state != gs_switch->state) {
		gs_switch->state = state;
		switch_set_state(&gs_switch->sdev, state);
	}

	schedule_delayed_work(&gs_switch->dwork, POLL_TIME);
}

static inline int sizeof_gs_switch_priv(int count)
{
	return sizeof(struct gs_switch_priv) +
		(sizeof(struct gs_switch_data) * count);
}

static int gs_switch_probe(struct platform_device *pdev)
{
	struct device_node *np = pdev->dev.of_node, *child;
	struct gs_switch_priv *priv;
	struct gs_switch_data *dev;
	int count, ret = 0;

	count = of_get_available_child_count(np);
	if (!count)
		return -ENODEV;

	priv = devm_kzalloc(&pdev->dev, sizeof_gs_switch_priv(count),
			    GFP_KERNEL);
	if (!priv)
		return -ENOMEM;

	for_each_available_child_of_node(np, child) {
		dev = &priv->devs[priv->num_devs++];

		dev->sdev.name = child->name;
		of_property_read_u32_array(child, "channel", &dev->channel_id, 1);
		of_property_read_u32_array(child, "voltage", dev->detect_voltage, 2);

		ret = switch_dev_register(&dev->sdev);
		if (ret < 0)
			goto err;

		INIT_DELAYED_WORK(&dev->dwork, gs_switch_work);

		schedule_delayed_work(&dev->dwork, POLL_TIME);
	}

	platform_set_drvdata(pdev, priv);

	return 0;
err:
	for (count = priv->num_devs - 2; count >= 0; count--) {
		switch_dev_unregister(&priv->devs[count].sdev);
		cancel_delayed_work_sync(&priv->devs[count].dwork);
	}

	return -ENODEV;
}

static int gs_switch_remove(struct platform_device *pdev)
{
	struct gs_switch_priv *priv = platform_get_drvdata(pdev);
	int i;

	for (i = 0; i < priv->num_devs; i++) {
		switch_dev_unregister(&priv->devs[i].sdev);
		cancel_delayed_work_sync(&priv->devs[i].dwork);
	}

	return 0;
}

static const struct of_device_id of_gs_switch_match[] = {
	{ .compatible = "grandstream,switch-adc", },
	{},
};

MODULE_DEVICE_TABLE(of, of_gs_switch_match);

static struct platform_driver gs_switch_driver = {
	.probe		 = gs_switch_probe,
	.remove		 = gs_switch_remove,
	.driver		 = {
		.name	 = "gs_switch_adc",
		.owner	 = THIS_MODULE,
		.of_match_table = of_match_ptr(of_gs_switch_match),
	},
};

module_platform_driver(gs_switch_driver);

MODULE_AUTHOR("Fanan Chen <fnchen@grandstream.cn>");
MODULE_DESCRIPTION("Grandstream EHS/PoE Switch driver");
MODULE_LICENSE("GPL");
