/*
 * Copyright (C) 2017  Grandstream Networks, Inc.
 * Author: Fanan Chen <fnchen@grandstream.cn>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/gpio.h>
#include <linux/interrupt.h>
#include <linux/of.h>
#include <linux/of_gpio.h>
#include <linux/of_platform.h>
#include <linux/platform_device.h>
#include <linux/slab.h>
#include <linux/workqueue.h>
#include <linux/switch.h>

struct gs_switch_data {
	int gpio;
	int irq;
	int active_low;
	struct work_struct work;
	struct switch_dev sdev;
};

struct gs_switch_priv {
	int num_devs;
	struct gs_switch_data devs[];
};

extern int fcfg_get_alarmin_reversal(unsigned short *reversal);

static void gs_switch_work(struct work_struct *work)
{
	int state;
	struct gs_switch_data *gs_switch =
		container_of(work, struct gs_switch_data, work);

	state = gpio_get_value(gs_switch->gpio);

	switch_set_state(&gs_switch->sdev,
			 gs_switch->active_low ? !state : state);
}

static irqreturn_t gs_switch_isr(int irq, void *data)
{
	struct gs_switch_data *gs_switch = (struct gs_switch_data *)data;

	schedule_work(&gs_switch->work);

	return IRQ_HANDLED;
}

static inline int sizeof_gs_switch_priv(int count)
{
	return sizeof(struct gs_switch_priv) +
		(sizeof(struct gs_switch_data) * count);
}

static int gs_switch_probe(struct platform_device *pdev)
{
	struct device_node *np = pdev->dev.of_node, *child;
	struct gs_switch_priv *priv;
	struct gs_switch_data *dev;
	int count, ret = 0;
	unsigned short reversal = 0;

	count = of_get_available_child_count(np);
	if (!count)
		return -ENODEV;

	priv = devm_kzalloc(&pdev->dev, sizeof_gs_switch_priv(count),
			    GFP_KERNEL);
	if (!priv)
		return -ENOMEM;

	fcfg_get_alarmin_reversal(&reversal);

	for_each_available_child_of_node(np, child) {
		enum of_gpio_flags flags;

		dev = &priv->devs[priv->num_devs++];

		dev->sdev.name = child->name;
		dev->gpio = of_get_gpio_flags(child, 0, &flags);
		if (reversal)
			dev->active_low = !(flags & OF_GPIO_ACTIVE_LOW);
		else
			dev->active_low = flags & OF_GPIO_ACTIVE_LOW;

		dev->irq = gpio_to_irq(dev->gpio);

		if (!gpio_is_valid(dev->gpio)) {
			dev_warn(&pdev->dev, "Skipping unavailable gpio%d (%s)\n",
				 dev->gpio, dev->sdev.name);
			continue;
		}

		ret = devm_gpio_request_one(&pdev->dev, dev->gpio, GPIOF_IN,
					   dev->sdev.name);
		if (ret < 0) {
			dev_err(&pdev->dev, "request gpio %d failed\n", dev->gpio);
			goto err;
		}

		ret = switch_dev_register(&dev->sdev);
		if (ret < 0)
			goto err;

		INIT_WORK(&dev->work, gs_switch_work);

		ret = request_irq(dev->irq, gs_switch_isr,
				  IRQF_TRIGGER_RISING | IRQF_TRIGGER_FALLING,
				  dev->sdev.name, dev);
		if (ret < 0) {
			switch_dev_unregister(&dev->sdev);
			goto err;
		}

		gs_switch_work(&dev->work);
	}

	platform_set_drvdata(pdev, priv);

	return 0;
err:
	for (count = priv->num_devs - 2; count >= 0; count--) {
		switch_dev_unregister(&priv->devs[count].sdev);
		cancel_work_sync(&priv->devs[count].work);
		free_irq(priv->devs[count].irq, &priv->devs[count]);
	}

	return -ENODEV;
}

static int gs_switch_remove(struct platform_device *pdev)
{
	struct gs_switch_priv *priv = platform_get_drvdata(pdev);
	int i;

	for (i = 0; i < priv->num_devs; i++) {
		switch_dev_unregister(&priv->devs[i].sdev);
		cancel_work_sync(&priv->devs[i].work);
		free_irq(priv->devs[i].irq, &priv->devs[i]);
	}

	return 0;
}

static const struct of_device_id of_gs_switch_match[] = {
	{ .compatible = "grandstream,switch-gpio", },
	{},
};

MODULE_DEVICE_TABLE(of, of_gs_switch_match);

static struct platform_driver gs_switch_driver = {
	.probe		 = gs_switch_probe,
	.remove		 = gs_switch_remove,
	.driver		 = {
		.name	 = "gs_switch",
		.owner	 = THIS_MODULE,
		.of_match_table = of_match_ptr(of_gs_switch_match),
	},
};

module_platform_driver(gs_switch_driver);

MODULE_AUTHOR("Fanan Chen <fnchen@grandstream.cn>");
MODULE_DESCRIPTION("Grandstream Camera/PoE Switch driver");
MODULE_LICENSE("GPL");
