/*
 * ak5703.c -- ASoC driver for AKM AK5703 ADC
 *
 * Copyright (C) 2015  Grandstream Networks, Inc.
 *
 * Author: Fanan Chen <fnchen@grandstream.cn>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/regmap.h>
#include <linux/regulator/consumer.h>
#include <linux/i2c.h>
#include <linux/gpio.h>
#include <linux/of_device.h>
#include <linux/of_gpio.h>
#include <linux/delay.h>
#include <sound/soc.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>
#include <sound/tlv.h>
#include <sound/ak5703.h>

#include "ak5703.h"

struct ak5703_priv {
	struct snd_soc_codec *codec;
	struct regulator *dvdd;
	struct regulator *avdd;
	struct regmap *regmap;
	unsigned int sysclk;
	unsigned int fmt;
	int master;
	int reset_gpio;
};

struct ak5703_rate_divs {
	u32 rate;
	u32 mclk;
	u8 cm;
	u8 fs;
};

static const struct reg_default ak5703_reg_defaults[] = {
	{0x00, 0x00}, {0x01, 0x18}, {0x02, 0x60}, {0x03, 0xC4},
	{0x04, 0x04}, {0x05, 0x0F}, {0x06, 0x00}, {0x07, 0x91},
	{0x08, 0x91}, {0x09, 0x84}, {0x0A, 0x00}, {0x0B, 0xE1},
	{0x0C, 0x00}, {0x0D, 0x00}, {0x0E, 0x00}, {0x0F, 0x00},

	{0x10, 0x00}, {0x11, 0x00}, {0x12, 0x60}, {0x13, 0x04},
	{0x14, 0x04}, {0x15, 0x00}, {0x16, 0x00}, {0x17, 0x91},
	{0x18, 0x91}, {0x19, 0x80}, {0x1A, 0x00}, {0x1B, 0xE1},
	{0x1C, 0x00}, {0x1D, 0x00}, {0x1E, 0x00}, {0x1F, 0x00},

	{0x20, 0xA9}, {0x21, 0x1F}, {0x22, 0xAD}, {0x23, 0x20},
	{0x24, 0x00}, {0x25, 0x00}, {0x26, 0x00}, {0x27, 0x00},
	{0x28, 0x00}, {0x29, 0x00}, {0x2A, 0x00}, {0x2B, 0x00},
	{0x2C, 0x00}, {0x2D, 0x00}, {0x2E, 0x00}, {0x2F, 0x00},

	{0x30, 0xA9}, {0x31, 0x1F}, {0x32, 0xAD}, {0x33, 0x20},
	{0x34, 0x00}, {0x35, 0x00}, {0x36, 0x00}, {0x37, 0x00},
};

static const struct ak5703_rate_divs ak5703_divs[] = {
	/* 8kHz rate */
	{8000, 2048000, 0x03, 0x00},
	{8000, 4096000, 0x01, 0x00},
	{8000, 8192000, 0x02, 0x00},
	/* 12kHz rate */
	{12000, 3072000, 0x03, 0x01},
	{12000, 6144000, 0x01, 0x01},
	{12000, 12288000, 0x02, 0x01},
	/* 16kHz rate */
	{16000, 4096000, 0x03, 0x02},
	{16000, 8192000, 0x01, 0x02},
	/* 24kHz rate */
	{24000, 6144000, 0x00, 0x03},
	{24000, 12288000, 0x01, 0x03},
	/* 11.025kHz rate */
	{11025, 2822400, 0x03, 0x05},
	{11025, 5644800, 0x01, 0x05},
	{11025, 11289600, 0x02, 0x05},
	/* 22.05kHz rate */
	{22050, 5644800, 0x03, 0x07},
	{22050, 11289600, 0x01, 0x07},
	/* 32kHz rate */
	{32000, 8192000, 0x00, 0x0A},
	/* 48kHz rate */
	{48000, 12288000, 0x00, 0x0B},
	{48000, 19200000, 0x00, 0x0B},
	/* 44.1kHz rate */
	{44100, 11289600, 0x00, 0x0F},
};

static const char *const ak5703_mic_input_mux[] = {
	"single-ended", "differential"
};

#define LINA_ENUM 0
#define RINA_ENUM 1
#define LINB_ENUM 2
#define RINB_ENUM 3

static const struct soc_enum ak5703_enum[] = {
	SOC_ENUM_SINGLE(AKM_MIC_CTRL_A, 0, 2, ak5703_mic_input_mux),
	SOC_ENUM_SINGLE(AKM_MIC_CTRL_A, 1, 2, ak5703_mic_input_mux),
	SOC_ENUM_SINGLE(AKM_MIC_CTRL_B, 0, 2, ak5703_mic_input_mux),
	SOC_ENUM_SINGLE(AKM_MIC_CTRL_B, 1, 2, ak5703_mic_input_mux),
};

/* Mic/Line Mux */
static const struct snd_kcontrol_new ak5703_lina_mux_control =
SOC_DAPM_ENUM("Route", ak5703_enum[LINA_ENUM]);
static const struct snd_kcontrol_new ak5703_rina_mux_control =
SOC_DAPM_ENUM("Route", ak5703_enum[RINA_ENUM]);
static const struct snd_kcontrol_new ak5703_linb_mux_control =
SOC_DAPM_ENUM("Route", ak5703_enum[LINB_ENUM]);
static const struct snd_kcontrol_new ak5703_rinb_mux_control =
SOC_DAPM_ENUM("Route", ak5703_enum[RINB_ENUM]);

static const struct snd_soc_dapm_widget ak5703_dapm_widgets[] = {
	SND_SOC_DAPM_MUX("Mic1 Input Mux", SND_SOC_NOPM, 0, 0,
			 &ak5703_lina_mux_control),
	SND_SOC_DAPM_MUX("Mic2 Input Mux", SND_SOC_NOPM, 0, 0,
			 &ak5703_rina_mux_control),
	SND_SOC_DAPM_MUX("Mic3 Input Mux", SND_SOC_NOPM, 0, 0,
			 &ak5703_linb_mux_control),
	SND_SOC_DAPM_MUX("Mic4 Input Mux", SND_SOC_NOPM, 0, 0,
			 &ak5703_rinb_mux_control),

	SND_SOC_DAPM_INPUT("MIC1"),
	SND_SOC_DAPM_INPUT("MIC2"),
	SND_SOC_DAPM_INPUT("MIC3"),
	SND_SOC_DAPM_INPUT("MIC4"),
};

static const struct snd_soc_dapm_route ak5703_dapm_routes[] = {
	{"Mic1 Input Mux", "single-ended", "MIC1"},
	{"Mic1 Input Mux", "differential", "MIC1"},

	{"Mic2 Input Mux", "single-ended", "MIC2"},
	{"Mic2 Input Mux", "differential", "MIC2"},

	{"Mic3 Input Mux", "single-ended", "MIC3"},
	{"Mic3 Input Mux", "differential", "MIC3"},

	{"Mic4 Input Mux", "single-ended", "MIC4"},
	{"Mic4 Input Mux", "differential", "MIC4"},

};

/* Mic gain amplifier volumes. From 0 to 36 dB */
static const unsigned int mic_amp_tlv[] = {
	TLV_DB_RANGE_HEAD(3),
	0, 1, TLV_DB_SCALE_ITEM(0, 800, 0),
	2, 4, TLV_DB_SCALE_ITEM(1200, 300, 0),
	5, 7, TLV_DB_SCALE_ITEM(2400, 600, 0),
};

/* Mic gain/sensitivity adjust. From -3 to +3 dB in 0.75 dB steps */
static const DECLARE_TLV_DB_SCALE(mic_gain_tlv, -300, 75, 0);
/* Mic Input Volume adjust. From -52.5 to +36 dB in 0.375 dB steps */
static const DECLARE_TLV_DB_SCALE(mic_input_tlv, -52500, 375, 0);

static const struct snd_kcontrol_new ak5703_snd_controls[] = {
	SOC_DOUBLE_R_TLV("Capture Volume", AKM_MIC_CTRL_A, AKM_MIC_CTRL_B,
			 4, 7, 0, mic_amp_tlv),
	SOC_SINGLE_TLV("Mic 1 Gain Volume", AKM_MIC_GAIN_A0, 0, 8, 0,
		       mic_gain_tlv),
	SOC_SINGLE_TLV("Mic 2 Gain Volume", AKM_MIC_GAIN_A1, 0, 8, 0,
		       mic_gain_tlv),
	SOC_SINGLE_TLV("Mic 3 Gain Volume", AKM_MIC_GAIN_B0, 0, 8, 0,
		       mic_gain_tlv),
	SOC_SINGLE_TLV("Mic 4 Gain Volume", AKM_MIC_GAIN_B1, 0, 8, 0,
		       mic_gain_tlv),

#if 0
	/* Power controls */
	SOC_SINGLE("Mic 1 Switch", AKM_PM_CTRL_A, 0, 0x01, 0),
	SOC_SINGLE("Mic 2 Switch", AKM_PM_CTRL_A, 1, 0x01, 0),
	SOC_SINGLE("Mic 3 Switch", AKM_PM_CTRL_B, 0, 0x01, 0),
	SOC_SINGLE("Mic 4 Switch", AKM_PM_CTRL_B, 1, 0x01, 0),
#endif

	/* Digital Volume controls */
	SOC_SINGLE_TLV("Mic 1 Input Volume", AKM_IVAL, 0, 0xF1, 0,
		       mic_input_tlv),
	SOC_SINGLE_TLV("Mic 2 Input Volume", AKM_IVAR, 0, 0xF1, 0,
		       mic_input_tlv),
	SOC_SINGLE_TLV("Mic 3 Input Volume", AKM_IVBL, 0, 0xF1, 0,
		       mic_input_tlv),
	SOC_SINGLE_TLV("Mic 4 Input Volume", AKM_IVBR, 0, 0xF1, 0,
		       mic_input_tlv),
};

static int ak5703_set_bias_level(struct snd_soc_codec *codec,
				 enum snd_soc_bias_level level)
{
//	struct ak5703_priv *ak5703 = snd_soc_codec_get_drvdata(codec);
	int ret = 0;

	switch (level) {
	case SND_SOC_BIAS_ON:
	case SND_SOC_BIAS_PREPARE:
		break;
	case SND_SOC_BIAS_STANDBY:
		if (codec->dapm.bias_level == SND_SOC_BIAS_OFF) {
#if 0
			if (ak5703->reset_gpio != -1)
				gpio_set_value(ak5703->reset_gpio, 1);
			mdelay(1);
			codec->cache_only = 0;
			/* execute a dummy command to ensure that all internal
			 * registers reset to their initial value*/
			snd_soc_write(codec, 0x00, 0x00);
			ret = snd_soc_cache_sync(codec);
			if (ret) {
				dev_err(codec->dev,
					"Failed to sync cache: %d\n", ret);
				return ret;
			}
			/* Enable Mic power */
			snd_soc_update_bits(codec, AKM_PM_CTRL_A, 0x08, 0x08);
			snd_soc_update_bits(codec, AKM_PM_CTRL_B, 0x08, 0x08);
#endif
		}
		break;
	case SND_SOC_BIAS_OFF:
#if 0
		if (ak5703->reset_gpio != -1)
			gpio_set_value(ak5703->reset_gpio, 0);
		mdelay(1);
		codec->cache_sync = 1;
		codec->cache_only = 1;
#endif
		break;
	}

	codec->dapm.bias_level = level;

	return ret;
}

static int ak5703_set_dai_sysclk(struct snd_soc_dai *codec_dai,
				 int clk_id, unsigned int freq, int dir)
{
	struct snd_soc_codec *codec = codec_dai->codec;
	struct ak5703_priv *ak5703 = snd_soc_codec_get_drvdata(codec);

	switch (freq) {
	case 2048000:
	case 2822400:
	case 3072000:
	case 4096000:
	case 5644800:
	case 6144000:
	case 8192000:
	case 11289600:
	case 12288000:
	case 19200000:
		ak5703->sysclk = freq;
		return 0;
	}

	return -EINVAL;
}

static int ak5703_set_dai_fmt(struct snd_soc_dai *codec_dai, unsigned int fmt)
{
	struct snd_soc_codec *codec = codec_dai->codec;
	struct ak5703_priv *ak5703 = snd_soc_codec_get_drvdata(codec);
	u8 pll_areg;

	pll_areg = snd_soc_read(codec, AKM_PLL_CTRL_A);

	/* Make AK5703 work in EXT mode */
	pll_areg &= ~PLL_MODE_SEL;

	/* set master/slave mode */
	switch (fmt & SND_SOC_DAIFMT_MASTER_MASK) {
	case SND_SOC_DAIFMT_CBM_CFM:
		pll_areg |= PLL_MASTER_MODE;
		ak5703->master = 1;
		break;
	case SND_SOC_DAIFMT_CBS_CFS:
		pll_areg &= ~PLL_MASTER_MODE;
		ak5703->master = 0;
		break;
	default:
		return -EINVAL;
	}

	/* set interface format */
	ak5703->fmt = fmt;

	/* update registers */
	snd_soc_write(codec, AKM_PLL_CTRL_A, pll_areg);

	return 0;
}

static u8 ak5703_get_dai_fmt(struct snd_soc_codec *codec,
			      struct snd_pcm_hw_params *params, unsigned int fmt) {
	struct ak5703_priv *ak5703 = snd_soc_codec_get_drvdata(codec);
	int wl = snd_pcm_format_width(params_format(params));
	/* bit[0:3] for Audio Interface Format
	 * bit[4:5] for BICK Output Frequency at Master Mode */
	u8 dai_fmt = 0;

	switch (fmt & SND_SOC_DAIFMT_FORMAT_MASK) {
	case SND_SOC_DAIFMT_I2S:
		if (wl == 16) {
			dai_fmt = 0x01;
		} else {
			dai_fmt = 0x03;
			if (ak5703->master)
				dai_fmt |= 1 << 4; /* BCLK = 64fs */
		}
		break;
	case SND_SOC_DAIFMT_LEFT_J:
		if (wl == 16) {
			dai_fmt = 0x00;
		} else {
			dai_fmt = 0x02;
			if (ak5703->master)
				dai_fmt |= 1 << 4; /* BCLK = 64fs */
		}
		break;
	case SND_SOC_DAIFMT_DSP_A:
		if (wl == 16) {
			dai_fmt = 0x07;
			if (ak5703->master)
				dai_fmt |= 1 << 4; /* BCLK = 64fs */
		} else {
			dai_fmt = 0x0F;
			if (ak5703->master)
				dai_fmt |= 2 << 4; /* BCLK = 128fs */
		}
		break;
	case SND_SOC_DAIFMT_DSP_B:
		if (wl == 16) {
			dai_fmt = 0x06;
			if (ak5703->master)
				dai_fmt |= 1 << 4; /* BCLK = 64fs */
		} else {
			dai_fmt = 0x0E;
			if (ak5703->master)
				dai_fmt |= 2 << 4; /* BCLK = 128fs */
		}
		break;
	}

	return dai_fmt;
}

static int ak5703_hw_params(struct snd_pcm_substream *substream,
			    struct snd_pcm_hw_params *params,
			    struct snd_soc_dai *codec_dai)
{
	struct snd_soc_codec *codec = codec_dai->codec;
	struct ak5703_priv *ak5703 = snd_soc_codec_get_drvdata(codec);
	int i, match = -1;
	u8 tdm, dif, fs_select, dai_fmt;

	dai_fmt = ak5703_get_dai_fmt(codec, params, ak5703->fmt);

	dif = (dai_fmt & 0x3) << 6;
	tdm = (((dai_fmt >> 2) & 0x3) << 6) | (dai_fmt & 0x30);

	switch (params_rate(params)) {
	case 8000:
		fs_select = 0x00;
		break;
	case 12000:
		fs_select = 0x01;
		break;
	case 16000:
		fs_select = 0x02;
		break;
	case 24000:
		fs_select = 0x03;
		break;
	case 11025:
		fs_select = 0x05;
		break;
	case 22050:
		fs_select = 0x07;
		break;
	case 32000:
		fs_select = 0x0A;
		break;
	case 48000:
		fs_select = 0x0B;
		break;
	case 44100:
		fs_select = 0x0F;
		break;
	default:
		return -EINVAL;
	}

	/* PLL configuration */
	for (i = 0; i < ARRAY_SIZE(ak5703_divs); i++) {
		if (ak5703_divs[i].rate == params_rate(params) &&
		    ak5703_divs[i].mclk == ak5703->sysclk) {
			match = i;
			break;
		}
	}

	if (match == -1) {
		dev_err(codec->dev,
			"Sample rate (%u) and format not supported\n",
			params_rate(params));
		return -EINVAL;
	}

	i = match;

	snd_soc_update_bits(codec, AKM_MIC_GAIN_A0, 0xC0, dif);
	snd_soc_update_bits(codec, AKM_CLKOUT_SELECT, 0x30,
			    ak5703_divs[i].cm << 4);
	snd_soc_update_bits(codec, AKM_FS_FILTER_CTRL_A, 0x0F,
			    ak5703_divs[i].fs);
	snd_soc_update_bits(codec, AKM_MIC_GAIN_A1, 0xF0, tdm);
	snd_soc_update_bits(codec, AKM_FS_FILTER_CTRL_A, 0x0F, fs_select);
	/* power up VCOM */
	snd_soc_update_bits(codec, AKM_PM_CTRL_A, 0x04, 0x04);

	/* Enable Mic power */
	snd_soc_update_bits(codec, AKM_PM_CTRL_A, 0x08, 0x08);
	snd_soc_update_bits(codec, AKM_PM_CTRL_B, 0x08, 0x08);
	mdelay(50);
	/* Enable ADC */
	snd_soc_update_bits(codec, AKM_PM_CTRL_A, 0x03, 0x03);
	snd_soc_update_bits(codec, AKM_PM_CTRL_B, 0x03, 0x03);
	mdelay(24);

	return 0;
}

static int ak5703_mute(struct snd_soc_dai *dai, int mute)
{
	struct snd_soc_codec *codec = dai->codec;

	if (mute) {
		snd_soc_update_bits(codec, AKM_PM_CTRL_A, 0x03, 0x00);
		snd_soc_update_bits(codec, AKM_PM_CTRL_B, 0x03, 0x00);
	} else {
		snd_soc_update_bits(codec, AKM_PM_CTRL_A, 0x03, 0x03);
		snd_soc_update_bits(codec, AKM_PM_CTRL_B, 0x03, 0x03);
	}

	return 0;
}

static int ak5703_init(struct snd_soc_codec *codec)
{
	struct ak5703_priv *ak5703 = snd_soc_codec_get_drvdata(codec);

	gpio_set_value(ak5703->reset_gpio, 0);
	/* "L" time of 1μs or more is needed to reset the AK5703 */
	udelay(2);
	gpio_set_value(ak5703->reset_gpio, 1);
	mdelay(1);
	/* execute a dummy command to ensure that all internal
	 * registers reset to their initial value*/
	snd_soc_write(codec, AKM_PM_CTRL_A, 0x00);

	snd_soc_write(codec, AKM_PM_CTRL_A, 0x00);
	snd_soc_write(codec, AKM_PM_CTRL_B, 0x00);

	snd_soc_write(codec, AKM_PLL_CTRL_A, 0x14);
	snd_soc_write(codec, AKM_MIC_CTRL_A, 0xD3);
	snd_soc_write(codec, AKM_MIC_GAIN_A0, 0x84);
	snd_soc_write(codec, AKM_MIC_GAIN_A1, 0x44);
	snd_soc_write(codec, AKM_FS_FILTER_CTRL_A, 0x8B);
	snd_soc_write(codec, AKM_CLKOUT_SELECT, 0x00);
	snd_soc_write(codec, 0x07, 0x91);
	snd_soc_write(codec, 0x08, 0x91);
	snd_soc_write(codec, 0x09, 0x04);
	snd_soc_write(codec, 0x0A, 0x00);
	snd_soc_write(codec, 0x0B, 0xE1);
	snd_soc_write(codec, 0x0C, 0x00);

	snd_soc_write(codec, 0x0D, 0x00);
	snd_soc_write(codec, 0x0E, 0x00);
	snd_soc_write(codec, 0x0F, 0x00);
	snd_soc_write(codec, 0x11, 0x00);
	snd_soc_write(codec, AKM_MIC_CTRL_B, 0xD3);
	snd_soc_write(codec, AKM_MIC_GAIN_B0, 0x04);
	snd_soc_write(codec, AKM_MIC_GAIN_B1, 0x04);
	snd_soc_write(codec, 0x15, 0x80);
	snd_soc_write(codec, 0x16, 0x00);
	snd_soc_write(codec, 0x17, 0x91);
	snd_soc_write(codec, 0x18, 0x91);
	snd_soc_write(codec, 0x19, 0x00);

	snd_soc_write(codec, 0x1A, 0x00);
	snd_soc_write(codec, 0x1B, 0xE1);
	snd_soc_write(codec, 0x1C, 0x00);
	snd_soc_write(codec, 0x1D, 0x00);
	snd_soc_write(codec, 0x1E, 0x00);
	snd_soc_write(codec, 0x1F, 0x00);
	snd_soc_write(codec, 0x20, 0xA9);
	snd_soc_write(codec, 0x21, 0x1F);
	snd_soc_write(codec, 0x22, 0xAD);
	snd_soc_write(codec, 0x23, 0x20);
	snd_soc_write(codec, 0x24, 0x3D);
	snd_soc_write(codec, 0x25, 0x19);
	snd_soc_write(codec, 0x26, 0x7A);
	snd_soc_write(codec, 0x27, 0x12);
	snd_soc_write(codec, 0x28, 0x00);
	snd_soc_write(codec, 0x29, 0x00);
	snd_soc_write(codec, 0x2A, 0x00);
	snd_soc_write(codec, 0x2B, 0x00);
	snd_soc_write(codec, 0x2C, 0x00);
	snd_soc_write(codec, 0x2D, 0x00);
	snd_soc_write(codec, 0x2E, 0x00);
	snd_soc_write(codec, 0x2F, 0x00);
	snd_soc_write(codec, 0x30, 0xA9);
	snd_soc_write(codec, 0x31, 0x1F);
	snd_soc_write(codec, 0x32, 0xAD);
	snd_soc_write(codec, 0x33, 0x20);
	snd_soc_write(codec, 0x34, 0x3D);
	snd_soc_write(codec, 0x35, 0x19);
	snd_soc_write(codec, 0x36, 0x7A);
	snd_soc_write(codec, 0x37, 0x12);
	snd_soc_write(codec, AKM_PLL_CTRL_A, 0x14);
	snd_soc_write(codec, AKM_MIC_GAIN_A0, 0x84);
	snd_soc_write(codec, AKM_FS_FILTER_CTRL_A, 0xAB);
	snd_soc_write(codec, 0x15, 0xA0);

	snd_soc_write(codec, AKM_PM_CTRL_A, 0x04);
	snd_soc_write(codec, AKM_PLL_CTRL_A, 0x14);
	snd_soc_write(codec, AKM_PM_CTRL_A, 0x0C);
	snd_soc_write(codec, AKM_PM_CTRL_B, 0x08);
	snd_soc_write(codec, AKM_MIC_CTRL_A, 0xD3);
	snd_soc_write(codec, 0x07, 0x91);
	snd_soc_write(codec, 0x08, 0x91);
	snd_soc_write(codec, AKM_MIC_CTRL_B, 0xD3);
	snd_soc_write(codec, 0x17, 0x91);
	snd_soc_write(codec, 0x18, 0x91);
	snd_soc_write(codec, AKM_PM_CTRL_A, 0x0F);
	snd_soc_write(codec, AKM_PM_CTRL_B, 0x0B);

	return 0;
}

static int ak5703_power_on(struct snd_soc_codec *codec, int on)
{
	struct ak5703_priv *ak5703 = snd_soc_codec_get_drvdata(codec);
	int ret = 0;

	if (on) {
		regulator_set_voltage(ak5703->dvdd, 1800000, 1800000);
		ret = regulator_enable(ak5703->dvdd);
		if (ret != 0) {
			dev_err(codec->dev, "Enable DVDD failed:%d\n", ret);
			return ret;
		}

		regulator_set_voltage(ak5703->avdd, 3300000, 3300000);
		ret = regulator_enable(ak5703->avdd);
		if (ret != 0) {
			dev_err(codec->dev, "Enable AVDD failed:%d\n", ret);
			return ret;
		}

	} else {
		ret = regulator_disable(ak5703->dvdd);
		if (ret != 0) {
			dev_err(codec->dev, "Disable DVDD failed:%d\n", ret);
			return ret;
		}

		ret = regulator_disable(ak5703->avdd);
		if (ret != 0) {
			dev_err(codec->dev, "Disable AVDD failed:%d\n", ret);
			return ret;
		}

	}

	return ret;
}

static int ak5703_startup(struct snd_pcm_substream *substream,
			  struct snd_soc_dai *dai)
{
	struct snd_soc_codec *codec = dai->codec;

	ak5703_power_on(codec, 1);

	return ak5703_init(codec);
}

static void ak5703_shutdown(struct snd_pcm_substream *substream,
			  struct snd_soc_dai *dai)
{
	struct snd_soc_codec *codec = dai->codec;

	ak5703_power_on(codec, 0);
}

static int ak5703_probe(struct snd_soc_codec *codec)
{
	struct ak5703_priv *ak5703 = snd_soc_codec_get_drvdata(codec);

	ak5703->codec = codec;

	return 0;
}

static int ak5703_remove(struct snd_soc_codec *codec)
{
	return 0;
}

static const struct regmap_config ak5703_regmap = {
	.reg_bits = 8,
	.val_bits = 8,

	.max_register = 0x37,
	.cache_type = REGCACHE_RBTREE,
	.reg_defaults = ak5703_reg_defaults,
	.num_reg_defaults = ARRAY_SIZE(ak5703_reg_defaults),
};

static struct snd_soc_codec_driver soc_codec_ak5703 = {
	.set_bias_level = ak5703_set_bias_level,
	.idle_bias_off = true,
	.probe = ak5703_probe,
	.remove = ak5703_remove,
	.controls = ak5703_snd_controls,
	.num_controls = ARRAY_SIZE(ak5703_snd_controls),
	.dapm_widgets = ak5703_dapm_widgets,
	.num_dapm_widgets = ARRAY_SIZE(ak5703_dapm_widgets),
	.dapm_routes = ak5703_dapm_routes,
	.num_dapm_routes = ARRAY_SIZE(ak5703_dapm_routes),
};

#define AK5703_RATES (SNDRV_PCM_RATE_8000 |\
					  SNDRV_PCM_RATE_11025 |	\
					  SNDRV_PCM_RATE_16000 |    \
					  SNDRV_PCM_RATE_22050 |    \
					  SNDRV_PCM_RATE_32000 |    \
					  SNDRV_PCM_RATE_44100 |    \
					  SNDRV_PCM_RATE_48000)

#define AK5703_FORMATS (SNDRV_PCM_FMTBIT_S16_LE |\
						SNDRV_PCM_FMTBIT_S24_LE |\
						SNDRV_PCM_FMTBIT_S32_LE)

static const struct snd_soc_dai_ops ak5703_dai_ops = {
	.set_sysclk = ak5703_set_dai_sysclk,
	.set_fmt = ak5703_set_dai_fmt,
	.hw_params = ak5703_hw_params,
	.digital_mute = ak5703_mute,
	.startup = ak5703_startup,
	.shutdown = ak5703_shutdown,
};

static struct snd_soc_dai_driver ak5703_dai = {
	.name = "ak5703-adc",
	.capture = {
		    .stream_name = "Capture",
		    .channels_min = 1,
		    .channels_max = 4,
		    .rates = AK5703_RATES,
		    .formats = AK5703_FORMATS,
		    },
	.ops = &ak5703_dai_ops,
};

static int ak5703_i2c_probe(struct i2c_client *i2c,
			    const struct i2c_device_id *id)
{
	struct ak5703_priv *ak5703;
	struct ak5703_platform_data *pdata = dev_get_platdata(&i2c->dev);
	int ret = 0;

	ak5703 = devm_kzalloc(&i2c->dev,
			      sizeof(struct ak5703_priv), GFP_KERNEL);
	if (!ak5703)
		return -ENOMEM;

	ak5703->regmap = devm_regmap_init_i2c(i2c, &ak5703_regmap);
	if (IS_ERR(ak5703->regmap)) {
		ret = PTR_ERR(ak5703->regmap);
		dev_err(&i2c->dev, "Failed to init regmap: %d\n", ret);
		return ret;
	}

	if (pdata) {
		ak5703->reset_gpio = pdata->reset_gpio;
	} else if (i2c->dev.of_node) {
		ak5703->reset_gpio = of_get_named_gpio(i2c->dev.of_node,
						       "reset-gpio", 0);
	}

	if (gpio_is_valid(ak5703->reset_gpio))
		if (devm_gpio_request_one(&i2c->dev, ak5703->reset_gpio,
					  GPIOF_OUT_INIT_LOW, "AK5703 PDN"))
			ak5703->reset_gpio = -EINVAL;

	ak5703->dvdd = devm_regulator_get(&i2c->dev, "dvdd");
	if (IS_ERR(ak5703->dvdd)) {
		ret = PTR_ERR(ak5703->dvdd);
		dev_err(&i2c->dev, "Failed to get DVDD:%d\n", ret);
		return ret;
	}

	ak5703->avdd = devm_regulator_get(&i2c->dev, "avdd");
	if (IS_ERR(ak5703->avdd)) {
		ret = PTR_ERR(ak5703->avdd);
		dev_err(&i2c->dev, "Failed to get DVDD:%d\n", ret);
		return ret;
	}

	i2c_set_clientdata(i2c, ak5703);

	ret = snd_soc_register_codec(&i2c->dev, &soc_codec_ak5703,
				     &ak5703_dai, 1);

	return ret;
}

static int ak5703_i2c_remove(struct i2c_client *i2c)
{
	snd_soc_unregister_codec(&i2c->dev);

	return 0;
}

static const struct i2c_device_id ak5703_i2c_id[] = {
	{"ak5703", 0},
	{}
};

MODULE_DEVICE_TABLE(i2c, ak5703_i2c_id);

static const struct of_device_id ak5703_of_match[] = {
	{ .compatible = "asahi-kasei,ak5703", },
	{}
};
MODULE_DEVICE_TABLE(of, ak5703_of_match);

static struct i2c_driver ak5703_i2c_driver = {
	.probe = ak5703_i2c_probe,
	.remove = ak5703_i2c_remove,
	.id_table = ak5703_i2c_id,
	.driver = {
		   .name = "ak5703",
		   .owner = THIS_MODULE,
		   .of_match_table = ak5703_of_match,
	},
};

module_i2c_driver(ak5703_i2c_driver);

MODULE_DESCRIPTION("ASoC driver for AK5703");
MODULE_AUTHOR("Fanan Chen <fnchen@grandstream.cn>");
MODULE_LICENSE("GPL");
